/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "InternalFilterSession.h"

namespace xma {

#define XMA_FILTER_SESSION "FilterSession"

///////////////////////////////////////////////////////////////////////////////
InternalFilterSession::InternalFilterSession(IntSessionFactoryPtr sf, XmaFilterProperties* props) : FilterThread<>(sf, props->params, props->param_cnt, 0, 1) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession()");

  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Failed to create VPI memory!");
  }

  if (props->input.format != XMA_VPE_FMT_TYPE) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Input must be on device!");
  }
  if (((props->input.width % 2) != 0) || ((props->input.height % 2) != 0)) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Input size(%dx%d) must be multiple of 2!",
        props->input.width, props->input.height);
  }

  VpiProcInitParams params;
  memset(&params, 0, sizeof(VpiProcInitParams));
  params.dev_ctx         = m_sf->GetDevHandle();
  params.mem_ctx         = m_mem_ctx.get();
  params.width           = props->input.width;
  params.height          = props->input.height;
  params.input_format    = InternalFrameProperties::GetPixFmtName(props->input.format);
  params.input_sw_format = InternalFrameProperties::GetPixFmtName(props->input.sw_format);
  params.flag            = props->input.flags;
  params.nb_frames       = 1;

  XmaParameter* param;
  switch (props->hwfilter_type) {
  case XMA_DRAWBOX_FILTER_TYPE:
    params.processor_type = VPE_PROCESSOR_DRAWBOX;
    param                 = GetParameter(props->params, props->param_cnt, XMA_FILTER_DRAWBOX_PARAM_COLOR);
    if (param) {
      params.box_color = *static_cast<uint32_t*>(param->value);
    } else {
      params.box_color = 0xFF000000;
    }
    param = GetParameter(props->params, props->param_cnt, XMA_FILTER_DRAWBOX_PARAM_THICKNESS);
    if (param) {
      params.box_thickness = *static_cast<int32_t*>(param->value);
    } else {
      params.box_thickness = 2;
    }
    break;
  default:
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Invalid filter type!");
  }
  m_type = props->hwfilter_type;

  if (((props->output.width % 2) != 0) || ((props->output.height % 2) != 0)) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Output size(%dx%d) must be multiple of 2!",
        props->output.width, props->output.height);
  }

  if (props->output.format != XMA_VPE_FMT_TYPE) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Output must be on device!");
  }

  VpiProcHandle new_handle = nullptr;
  VpiRet        ret        = vpi_processor_init(&params, &new_handle);
  if (ret < VPI_SUCCESS) {
    Close();
    THROW_RUNTIME_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Failed to initialize filter: %s!", vpi_error_str(ret));
  }
  m_handle = filter_handle_t(new_handle, [](void* handle) { vpi_processor_close(handle); });

  if (vpi_processor_open(m_handle.get()) < 0) {
    m_handle.reset();
    Close();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() Unable to open filter for processing");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::InternalFilterSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFilterSession::~InternalFilterSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::~InternalFilterSession()");

  Close();

  m_handle.reset();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFilterSession::Enqueue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Enqueue()");

  int32_t ret = FilterThread::Push(data);

  switch (ret) {
  case XMA_TRY_AGAIN:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Enqueue() - Returning try again");
    break;
  case XMA_SUCCESS:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Enqueue() - Returning success");
    break;
  default:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Enqueue() - Returning");
    break;
  }
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFilterSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Dequeue()");

  int32_t ret = FilterThread::Pull(data);

  char error_msg[1024];
  sprintf(error_msg, "InternalFilterSession::Dequeue() - returning %d", ret);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, error_msg);

  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFilterSession::Process(queue_t& in, queue_t& out, EmptyDynamicParams*) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process()");
  FramePtr frame = std::move(std::get<QueueFrame>(in).frame);
  if (!frame) {
    out = QueueFrame();
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() - Returning EOS");
    return XMA_EOS;
  }

  using vpi_frm_t          = std::unique_ptr<VpiFrm, std::function<void(VpiFrm*)>>;
  IntSessionFactoryPtr sf  = m_sf;
  VpiFrm*              pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Out of memory!");
  }
  InitVpiFrm(pic);
  vpi_frm_t in_pic = vpi_frm_t(pic, [sf](VpiFrm* frm) {
    vpi_frame_unref(frm);
    if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf)) == 0) {
      vpi_frame_free(sf->GetMemHandle(), frm);
    }
  });

  int32_t ret;
  if ((ret = AddFrameToVpiFrm(m_sf, m_log, frame, in_pic.get())) != XMA_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Failed to convert XmaFrame to VpiFrm!");
  }

  pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Out of memory!");
  }
  InitVpiFrm(pic);
  vpi_frm_t out_pic = vpi_frm_t(pic, [sf](VpiFrm* frm) {
    vpi_frame_unref(frm);
    if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf)) == 0) {
      vpi_frame_free(sf->GetMemHandle(), frm);
    }
  });

  if (vpi_processor_process_frame(m_handle.get(), in_pic.get(), out_pic.get()) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Failed to process frame in filter!");
  }

  out = QueueFrame();

  switch (m_type) {
  case XMA_DRAWBOX_FILTER_TYPE: {
    in_pic.reset();
    out_pic->valid        = 1; // GC620 doesn't properly set this
    IntFramePtr out_frame = std::make_shared<InternalFrame>(m_sf, m_log, out_pic.release());
    if (!out_frame) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Out of memory!");
    }

    // To Do: Remove this once the filter passes this side data
    SideDataPtr sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS);
    if (sd && (!out_frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS))) {
      out_frame->AddSideData(sd);
    }

    std::get<QueueFrame>(out).frame = out_frame;
    break;
  }
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Invalid filter type!");
  }

  if (!std::get<QueueFrame>(out).frame) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() Out of memory!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FILTER_SESSION, "InternalFilterSession::Process() - Returning success");
  return XMA_SUCCESS;
}

} // namespace xma
