/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "xmaapi.h"
#include "InternalSideData.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"
#include <stdlib.h>
#include <cstring>

#define XMA_BUFFERS_MOD "xmabuffers"

///////////////////////////////////////////////////////////////////////////////
XmaFrame* xma_frame_alloc(XmaHandle handle, XmaFrameProperties* frame_props, bool dummy) {
  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  if (!xma_sf) {
    return nullptr;
  }

  if (!frame_props) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_alloc() Invalid parameter!");
    return nullptr;
  }
  if (!dummy && (frame_props->width > MAX_FRAME_W_H || frame_props->height > MAX_FRAME_W_H)) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_alloc() Frame size is above limits: w=%d, h=%d", frame_props->width, frame_props->height);
    return nullptr;
  }
  if (!dummy && (frame_props->width < 0 || frame_props->height < 0)) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_alloc() Frame size is invalid: w=%d, h=%d", frame_props->width, frame_props->height);
    return nullptr;
  }

  TRY()
  if (dummy) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_alloc(dummy)");
  } else {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_alloc(format=%s, sw_format=%s, size=%dx%d)",
        xma::InternalFrameProperties::GetPixFmtName(frame_props->format), xma::InternalFrameProperties::GetPixFmtName(frame_props->sw_format), frame_props->width,
        frame_props->height);
  }
  xma::IntFramePropertiesPtr props  = std::make_shared<xma::InternalFrameProperties>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), frame_props);
  xma::buffer_t              buffer = dummy ? nullptr : xma_sf->m_sf->GetBuffer(xma::InternalFrameProperties::GetBufferSize(props), XMA_HOST_BUFFER_TYPE);
  xma::IntFramePtr           frame  = std::make_shared<xma::InternalFrame>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), props, dummy, std::move(buffer));
  frame->SetSelf();
  XmaFrame* ret = frame->GetFrame();
  frame.reset();
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_alloc() - Returning");
  return ret;
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_planes_get(XmaHandle handle, XmaFrameProperties* frame_props) {
  if (!frame_props) {
    XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
    if (xma_sf) {
      xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_planes_get() Invalid parameter!");
    }
    return XMA_ERROR_INVALID;
  }
  return vpi_frame_get_planes(frame_props->sw_format);
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_get_plane_height(XmaHandle handle, XmaFrameProperties* frame_props, size_t plane) {
  if (!frame_props) {
    XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
    if (xma_sf) {
      xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_get_plane_height() Invalid parameter!");
    }
    return XMA_ERROR_INVALID;
  }
  return vpi_frame_get_plane_height(frame_props->sw_format, plane, frame_props->height, XMA_FRAME_H_ALIGN);
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_get_plane_stride(XmaHandle handle, XmaFrameProperties* frame_props, size_t plane) {
  if (!frame_props) {
    XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
    if (xma_sf) {
      xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_get_plane_stride() Invalid parameter!");
    }
    return XMA_ERROR_INVALID;
  }
  return vpi_frame_get_plane_stride(frame_props->sw_format, plane, frame_props->width, XMA_FRAME_W_ALIGN);
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_get_plane_size(XmaHandle handle, XmaFrameProperties* frame_props, size_t plane) {
  if (!frame_props) {
    XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
    if (xma_sf) {
      xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_get_plane_size() Invalid parameter!");
    }
    return XMA_ERROR_INVALID;
  }
  return vpi_frame_get_plane_size(frame_props->sw_format, plane, frame_props->width, XMA_FRAME_W_ALIGN, frame_props->height, XMA_FRAME_H_ALIGN);
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_inc_ref(XmaFrame* xma_frame) {
  if (!xma_frame) {
    return 0;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_inc_ref", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_inc_ref(format=%s, sw_format=%s, size=%dx%d)",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height);
  }
  int32_t ref_count = 0;
  if (frame) {
    ref_count = frame->IncRef();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_inc_ref() - Returning");
  }
  return ref_count;
  CATCH_RET(log, XMA_BUFFERS_MOD, 0)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_dec_ref(XmaFrame* xma_frame) {
  if (!xma_frame) {
    return 0;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_dec_ref", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_dec_ref(format=%s, sw_format=%s, size=%dx%d)",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height);
  }
  int32_t ref_count = 0;
  if (frame) {
    if (frame->GetRefCount() == 1) {
      info->SetSelf(nullptr);
    }
    ref_count = frame->DecRef();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_dec_ref() - Returning");
  }
  return ref_count;
  CATCH_RET(log, XMA_BUFFERS_MOD, 0)
}

///////////////////////////////////////////////////////////////////////////////
void xma_frame_free(XmaFrame* xma_frame) {
  if (!xma_frame) {
    return;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_free(format=%s, sw_format=%s, size=%dx%d)",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height);
  }
  if (frame) {
    if (frame->GetRefCount() == 1) {
      info->SetSelf(nullptr);
    } else {
      frame->DecRef();
    }
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_free() - Returning");
  }
  CATCH(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
XmaFrame* xma_frame_from_buffers_clone(
    XmaHandle handle, XmaFrameProperties* frame_props, XmaFrameData* frame_data, xma_frame_clone_free_callback_function free_callback, void* opaque) {
  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  if (!xma_sf) {
    return nullptr;
  }

  if ((!frame_props) || (!frame_data)) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_from_buffers_clone() Invalid parameter!");
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_from_buffers_clone(format=%s, sw_format=%s, size=%dx%d)",
      xma::InternalFrameProperties::GetPixFmtName(frame_props->format), xma::InternalFrameProperties::GetPixFmtName(frame_props->sw_format), frame_props->width,
      frame_props->height);
  xma::IntFramePropertiesPtr props = std::make_shared<xma::InternalFrameProperties>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), frame_props);
  xma::IntFramePtr frame = std::make_shared<xma::InternalFrame>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), props, frame_data->data, [free_callback, opaque](uint8_t** data) {
    if (free_callback) {
      XmaFrameData callback_data;
      for (int32_t plane = 0; plane < XMA_MAX_PLANES; ++plane) {
        callback_data.data[plane] = data[plane];
      }
      free_callback(opaque, &callback_data);
    }
  });
  frame->SetSelf();
  XmaFrame* ret = frame->GetFrame();
  frame.reset();
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_from_buffers_clone() - Returning");
  return ret;
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
XmaFrame* xma_frame_clone(XmaHandle handle, XmaFrame* xma_frame) {
  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  if (!xma_sf) {
    return nullptr;
  }

  if (!xma_frame) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_clone() Invalid parameter!");
    return nullptr;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_clone(format=%s, sw_format=%s, size=%dx%d)",
      xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
      xma_frame->frame_props.width, xma_frame->frame_props.height);
  xma::IntFramePtr new_frame = std::make_shared<xma::InternalFrame>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), frame);
  new_frame->SetSelf();
  XmaFrame* ret = new_frame->GetFrame();
  new_frame.reset();
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_clone() - Returning");
  return ret;
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideData* xma_side_data_alloc(XmaHandle handle, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size) {
  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_alloc(type=%d, buffer_type=%s, size=%lu)", type,
      ((buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (buffer_type == NO_BUFFER)                   ? "none" :
                                                             "unknown"),
      size);
  xma::IntSideDataPtr       sd   = std::make_shared<xma::InternalSideData>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), type, buffer_type, size);
  XmaFrameSideData*         ret  = sd->GetSideData();
  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(ret->private_do_not_touch);
  info->SetSelf(sd);
  sd.reset();
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_alloc() - Returning");
  return ret;
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
void xma_side_data_free(XmaFrameSideData* side_data) {
  if (!side_data) {
    return;
  }

  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::LogPtr               log;
  xma::IntSideDataPtr       sd;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    sd  = info->GetSelf();
  }

  TRY()
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_free(type=%d, buffer_type=%s, size=%lu)", side_data->type,
        ((side_data->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
                (side_data->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
                (side_data->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
                (side_data->buffer_type == NO_BUFFER)                   ? "none" :
                                                                          "unknown"),
        side_data->size);
  }
  if (sd) {
    if (sd->GetRefCount() == 1) {
      info->SetSelf(nullptr);
    }
    sd->DecRef();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_free() - Returning");
  }
  CATCH(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_side_data_inc_ref(XmaFrameSideData* side_data) {
  if (!side_data) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::LogPtr               log;
  xma::IntSideDataPtr       sd;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    sd  = info->GetSelf();
  }

  TRY()
  LOCK_SIDE_DATA(log, "xma_side_data_inc_ref", XMA_BUFFERS_MOD, side_data);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_inc_ref(type=%d)", static_cast<int>(side_data->type));
  }
  int32_t ref_count = 0;
  if (sd) {
    ref_count = sd->IncRef();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_inc_ref() - Returning");
  }
  return ref_count;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_side_data_dec_ref(XmaFrameSideData* side_data) {
  if (!side_data) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::LogPtr               log;
  xma::IntSideDataPtr       sd;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    sd  = info->GetSelf();
  }

  TRY()
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_dec_ref(type=%d)", static_cast<int>(side_data->type));
  }
  int32_t ref_count = 0;
  if (sd) {
    if (sd->GetRefCount() == 1) {
      info->SetSelf(nullptr);
    }
    ref_count = sd->DecRef();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_dec_ref() - Returning");
  }
  return ref_count;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_side_data_get_refcount(XmaFrameSideData* side_data) {
  if (!side_data) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::LogPtr               log;
  xma::IntSideDataPtr       sd;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    sd  = info->GetSelf();
  }

  TRY()
  LOCK_SIDE_DATA(log, "xma_side_data_get_refcount", XMA_BUFFERS_MOD, side_data);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_get_refcount(type=%d)", static_cast<int>(side_data->type));
  }
  int32_t ref_count = 0;
  if (sd) {
    ref_count = sd->GetRefCount();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_get_refcount() - Returning");
  }
  return ref_count;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_side_data_read(XmaFrameSideData* side_data) {
  if (!side_data) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::LogPtr               log;
  xma::IntSideDataPtr       sd;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    sd  = info->GetSelf();
  }

  if (side_data->buffer_type != XMA_DEVICE_BUFFER_TYPE) {
    if (log) {
      log->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_side_data_read() Invalid parameter");
    }

    return XMA_ERROR_INVALID;
  }

  TRY()
  LOCK_SIDE_DATA(log, "xma_side_data_read", XMA_BUFFERS_MOD, side_data);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_read(type=%d)", static_cast<int>(side_data->type));
  }
  int32_t read = 0;
  if (sd) {
    read = sd->Read();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_read() - Returning");
  }
  return read;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_side_data_write(XmaFrameSideData* side_data) {
  if (!side_data) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateSideDataInfo* info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::LogPtr               log;
  xma::IntSideDataPtr       sd;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    sd  = info->GetSelf();
  }

  if (side_data->buffer_type != XMA_DEVICE_BUFFER_TYPE) {
    if (log) {
      log->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_side_data_write() Invalid parameter");
    }

    return XMA_ERROR_INVALID;
  }

  TRY()
  LOCK_SIDE_DATA(log, "xma_side_data_write", XMA_BUFFERS_MOD, side_data);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_write(type=%d)", static_cast<int>(side_data->type));
  }
  int32_t written = 0;
  if (sd) {
    written = sd->Write();
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_side_data_write() - Returning");
  }
  return written;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_add_side_data(XmaFrame* xma_frame, XmaFrameSideData* side_data) {
  if ((!xma_frame) || (!side_data)) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }
  xma::PrivateSideDataInfo* sd_info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::IntSideDataPtr       sd;
  if ((sd_info) && (sd_info->IsValid())) {
    sd = sd_info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_add_side_data", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_add_side_data(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%d))",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height, static_cast<int>(side_data->type));
  }
  int32_t ret = XMA_ERROR_INVALID;
  if (frame && sd) {
    ret = frame->AddSideData(sd);
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_add_side_data() - Returning");
  }
  return ret;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideData* xma_frame_get_side_data(XmaFrame* xma_frame, enum XmaFrameSideDataType type) {
  if ((!xma_frame) || (type <= XMA_FRAME_SIDE_DATA_INVALID) || (XMA_FRAME_SIDE_DATA_MAX_COUNT <= type)) {
    return nullptr;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_get_side_data", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_get_side_data(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%d))",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height, static_cast<int>(type));
  }
  xma::SideDataPtr side_data = frame->GetSideData(type);
  if (!side_data) {
    if (log) {
      log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_get_side_data() - Returning NULL");
    }
    return nullptr;
  }

  xma::IntSideDataPtr sd = std::dynamic_pointer_cast<xma::InternalSideData>(side_data);
  if (sd) {
    sd->IncRef();
    XmaFrameSideData* ret = sd->GetSideData();
    if (log) {
      log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_get_side_data() - Returning");
    }
    return ret;
  } else {
    if (log) {
      log->LogMsg(XMA_ERROR_LOG, XMA_BUFFERS_MOD, "xma_frame_get_side_data() Invalid side data! Are you mixing the C and C++ interfaces?");
    }
    return nullptr;
  }
  CATCH_NULL(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_remove_side_data(XmaFrame* xma_frame, XmaFrameSideData* side_data) {
  if ((!xma_frame) || (!side_data)) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }
  xma::PrivateSideDataInfo* sd_info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch);
  xma::IntSideDataPtr       sd;
  if ((sd_info) && (sd_info->IsValid())) {
    sd = sd_info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_remove_side_data", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_remove_side_data(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%d))",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height, static_cast<int>(side_data->type));
  }

  int32_t ret = XMA_ERROR_INVALID;
  if (frame && sd) {
    ret = frame->RemoveSideData(sd);
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_remove_side_data() - Returning");
  }
  return ret;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_frame_remove_side_data_type(XmaFrame* xma_frame, enum XmaFrameSideDataType type) {
  if ((!xma_frame) || (type <= XMA_FRAME_SIDE_DATA_INVALID) || (XMA_FRAME_SIDE_DATA_MAX_COUNT <= type)) {
    return XMA_ERROR_INVALID;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_remove_side_data_type", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_remove_side_data_type(frame=(format=%s, sw_format=%s, size=%dx%d), type=%d)",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height, static_cast<int>(type));
  }
  int32_t ret = XMA_ERROR_INVALID;
  if (frame) {
    ret = frame->RemoveSideDataType(type);
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_remove_side_data_type() - Returning");
  }
  return ret;
  CATCH_ERROR(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
void xma_frame_clear_all_side_data(XmaFrame* xma_frame) {
  if (!xma_frame) {
    return;
  }

  xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
  xma::LogPtr            log;
  xma::IntFramePtr       frame;
  if ((info) && (info->IsValid())) {
    log   = info->GetLog();
    frame = info->GetSelf();
  }

  TRY()
  LOCK_XMA_FRAME(log, "xma_frame_clear_all_side_data", XMA_BUFFERS_MOD, xma_frame);

  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_clear_all_side_data(frame=(format=%s, sw_format=%s, size=%dx%d))",
        xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.format), xma::InternalFrameProperties::GetPixFmtName(xma_frame->frame_props.sw_format),
        xma_frame->frame_props.width, xma_frame->frame_props.height);
  }
  frame->ClearAllSideData();
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_frame_clear_all_side_data() - Returning");
  }
  CATCH(log, XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
XmaDataBuffer* xma_data_buffer_alloc(XmaHandle handle, size_t size, bool dummy) {
  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  if (dummy) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_buffer_alloc(dummy)");
  } else {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_buffer_alloc(size=%lu)", size);
  }
  xma::IntDataBufferPtr buf = std::make_shared<xma::InternalDataBuffer>(xma_sf->m_sf, xma_sf->m_sf->GetLog(), size, dummy);
  buf->SetSelf();
  XmaDataBuffer* ret = buf->GetDataBuffer();
  buf.reset();
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_buffer_alloc() - Returning");
  return ret;
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
XmaDataBuffer* xma_data_from_buffer_clone(XmaHandle handle, uint8_t* data, size_t size, xma_data_buffer_clone_free_callback_function free_callback, void* opaque) {
  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_from_buffer_clone(size=%lu)", size);
  xma::IntDataBufferPtr buf = std::make_shared<xma::InternalDataBuffer>(
      xma_sf->m_sf, xma_sf->m_sf->GetLog(), data, size, [free_callback, opaque](uint8_t* callback_data) {
        if (free_callback) {
          free_callback(opaque, callback_data);
        }
      });

  buf->SetSelf();
  XmaDataBuffer* ret = buf->GetDataBuffer();
  buf.reset();
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_from_buffer_clone() - Returning");
  return ret;
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_BUFFERS_MOD)
}

///////////////////////////////////////////////////////////////////////////////
void xma_data_buffer_free(XmaDataBuffer* data) {
  if (!data) {
    return;
  }

  xma::PrivateDataBufferInfo* info = static_cast<xma::PrivateDataBufferInfo*>(data->private_do_not_touch);
  xma::LogPtr                 log;
  xma::IntDataBufferPtr       buf;
  if ((info) && (info->IsValid())) {
    log = info->GetLog();
    buf = std::dynamic_pointer_cast<xma::InternalDataBuffer>(info->GetSelf());
  }

  TRY()
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_buffer_free(size=%d)", data->alloc_size);
  }
  if (buf) {
    if (buf->GetRefCount() == 1) {
      info->SetSelf(nullptr);
    } else {
      buf->DecRef();
    }
  }
  if (log) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFERS_MOD, "xma_data_buffer_free() - Returning");
  }
  CATCH(log, XMA_BUFFERS_MOD)
}
